import functools
import inspect
import os
import subprocess
from enum import Enum

import psutil

from mlflow_extensions.log import Logger, get_logger

LOGGER: Logger = get_logger()


class Modality(Enum):
    VISION = "vision"
    TEXT = "text"
    AUDIO = "audio"


def run_if(modality: Modality):
    def outer(func):

        @functools.wraps(func)
        def inner(*args, **kwargs):
            this_modality = kwargs.get("modality_type")
            if this_modality is None:
                raise ValueError("modality_type must be provided")
            if modality == this_modality:
                return func(*args, **kwargs)

            print(
                f"Skipping {func.__name__} because modality is {this_modality} not {modality}"
            )
            return

        return inner

    return outer


def inject_openai_client(func):
    # error if function has regular args
    has_args = inspect.getfullargspec(func).args
    if has_args:
        raise ValueError("Function cannot have regular arguments")

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        host = kwargs.get("host", "0.0.0.0")
        port = kwargs.get("port", 9989)
        from openai import OpenAI

        client = OpenAI(base_url=f"http://{host}:{port}/v1", api_key="local")
        return func(*args, client=client, **kwargs)

    return wrapper


class ServerFramework(Enum):
    VLLM = "vllm"
    SGLANG = "sglang"


def get_process_ids(search_string):
    try:
        # Run 'ps aux' to get the process list
        result = subprocess.run(
            ["ps", "aux"], text=True, capture_output=True, check=True
        )

        # Process the output
        pids = []
        for line in result.stdout.splitlines():
            if search_string in line:
                # Extract PID (assumed to be in the second column)
                LOGGER.info(f"Found orphaned process: {line} matching {search_string}")
                parts = line.split()
                if len(parts) > 1 and parts[1].isdigit():
                    pids.append(parts[1])

        return pids

    except subprocess.CalledProcessError as e:
        print(f"An error occurred: {e}")
        return []


def kill_process(pid):
    try:
        subprocess.run(["kill", "-9", pid], check=True)
        print(f"Killed process with PID {pid}")
    except subprocess.CalledProcessError as e:
        print(f"Failed to kill process with PID {pid}: {e}")


def kill_processes_containing(search_string):
    pids = get_process_ids(search_string)
    for pid in pids:
        kill_process(pid)


def is_process_active(pid: int) -> bool:
    try:
        process = psutil.Process(pid)
        status = process.status()
        if status == psutil.STATUS_ZOMBIE:
            print(f"Process {pid} is a zombie.")
            return False
        else:
            print(f"Process {pid} is running with status: {status}.")
            return True
    except psutil.NoSuchProcess:
        print(f"Process {pid} does not exist (killed or never existed).")
        return False
    except psutil.AccessDenied:
        print(f"Access denied to process {pid}.")
        return False



import functools
from functools import wraps
from typing import Optional, Type
from urllib.parse import urlparse

import httpx
from httpx import URL, AsyncClient, Client, Request, Response

from mlflow_extensions.serving.serde_v2 import MlflowPyfuncHttpxSerializer


def is_local(url: str) -> bool:
    return "0.0.0.0" in url or "localhost" in url or "127.0.0.1" in url


def validate_url_token(url: str, token: Optional[str] = None) -> bool:
    if is_local(url) is False and token is None:
        return False
    return True


def build_endpoint_url(url: str) -> str:
    normalized_url = url.rstrip("/")
    if normalized_url.endswith("/invocations"):
        normalized_url = normalized_url.replace("/invocations", "")
    if is_local(normalized_url):
        return normalized_url.rstrip("/")
    return normalized_url


def get_ezdeploy_lite_openai_url(
    model_deployment_name: str,
    databricks_host: str = None,
    databricks_token: str = None,
    prefix: str = None,
):
    from mlflow_extensions.databricks.deploy.ez_deploy_lite import (
        EZ_DEPLOY_LITE_PREFIX,
        EzDeployLiteManager,
    )

    edlm = EzDeployLiteManager(
        databricks_host=databricks_host,
        databricks_token=databricks_token,
        prefix=prefix or EZ_DEPLOY_LITE_PREFIX,
    )
    return edlm.get_openai_url(model_deployment_name)


class BaseCustomMLFlowHttpClient:
    def __init__(
        self,
        *,
        endpoint_url: str,
        token: Optional[str] = None,
        timeout: int = 30,
        requires_openai_compat: bool = True,
    ):
        if validate_url_token(endpoint_url, token) is False:
            raise ValueError(
                "You must provide a token unless the endpoint is localhost or 0.0.0.0"
            )

        headers = {
            "Content-Type": "application/json",
        }
        if token:
            headers["Authorization"] = f"Bearer {token}"

        self._custom_provided_base_path = urlparse(endpoint_url).path.rstrip("/")
        base_url = build_endpoint_url(endpoint_url)
        self._timeout = timeout
        self._http_client_args = {
            "base_url": base_url,
            "headers": headers,
            "timeout": timeout,
        }
        self.requires_openai_compat = requires_openai_compat

    def _prepare_request(self, request: Request) -> dict:
        url = request.url
        if isinstance(request.url, str):
            url = httpx.URL(request.url)
        path_to_request = url.path.replace(self._custom_provided_base_path, "")
        return {
            "inputs": [
                MlflowPyfuncHttpxSerializer.serialize_request(
                    request,
                    path_to_request,
                    requires_openai_compat=self.requires_openai_compat,
                )
            ]
        }

    @staticmethod
    def _process_response(response_json: dict, orig_request: Request) -> Response:
        try:
            prediction = response_json["predictions"][0]
        except KeyError:
            raise ValueError(
                f"Invalid response from server: missing 'predictions' key; response_json={response_json}"
            )
        return MlflowPyfuncHttpxSerializer.deserialize_response(
            prediction, orig_request
        )


class CustomMLFlowHttpClient(BaseCustomMLFlowHttpClient, Client):
    def __init__(self, **kwargs):
        BaseCustomMLFlowHttpClient.__init__(self, **kwargs)
        Client.__init__(self)
        self._http_client = httpx.Client(**self._http_client_args)

    def send(self, request: Request, **kwargs) -> Response:
        inputs = self._prepare_request(request)
        response = self._http_client.post("/invocations", json=inputs)
        return self._process_response(response.json(), request)


class AsyncCustomMLFlowHttpClient(BaseCustomMLFlowHttpClient, AsyncClient):
    def __init__(self, **kwargs):
        BaseCustomMLFlowHttpClient.__init__(self, **kwargs)
        AsyncClient.__init__(self)
        self._http_client = httpx.AsyncClient(**self._http_client_args)

    async def send(self, request: Request, **kwargs) -> Response:
        inputs = self._prepare_request(request)
        response = await self._http_client.post("/invocations", json=inputs)
        return self._process_response(response.json(), request)


def inject_mlflow_openai_compat_client(
    use_sync: Optional[bool] = None,
    use_async: Optional[bool] = None,
    sync_client_attribute: str = "http_client",
    async_client_attribute: str = "async_client",
):
    def decorator(cls):
        original_init = cls.__init__

        @wraps(original_init)
        def new_init(self, **kwargs):
            if kwargs.get("streaming") is True:
                raise ValueError("Streaming is not supported with this adapter")
            assert "base_url" in kwargs, "You must provide a base_url"
            base_url = kwargs.get("base_url")
            api_key = kwargs.get("api_key")
            timeout = int(kwargs.get("timeout", 30))
            if validate_url_token(base_url, api_key) is False:
                raise ValueError("You must provide an api_key")

            additional_client_kwargs = {}
            if use_sync is True:
                additional_client_kwargs[sync_client_attribute] = (
                    CustomMLFlowHttpClient(
                        endpoint_url=base_url, token=api_key, timeout=timeout
                    )
                )

            if use_async is True:
                additional_client_kwargs[async_client_attribute] = (
                    AsyncCustomMLFlowHttpClient(
                        endpoint_url=base_url, token=api_key, timeout=timeout
                    )
                )

            kwargs.update(additional_client_kwargs)
            original_init(self, **kwargs)

        cls.__init__ = new_init
        return cls

    return decorator


CustomServerClient = functools.partial(  # noqa
    CustomMLFlowHttpClient, requires_openai_compat=False
)
CustomServerClient: Type[BaseCustomMLFlowHttpClient]
CustomServerAsyncClient = functools.partial(  # noqa
    AsyncCustomMLFlowHttpClient, requires_openai_compat=False
)
CustomServerAsyncClient: Type[BaseCustomMLFlowHttpClient]



import datetime
import logging
import os
from datetime import datetime
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple

import structlog
from databricks.sdk import WorkspaceClient
from structlog.stdlib import BoundLogger
from tenacity import retry, stop_after_attempt, wait_fixed

logger: BoundLogger = structlog.get_logger(__name__)


def full_volume_name_to_path(full_volume_name: str) -> Optional[str]:
    if full_volume_name is None:
        return None

    catalog_name: str
    schema_name: str
    volume_name: str
    parts: List[str] = full_volume_name.split(".")
    if len(parts) != 3:
        return None

    catalog_name, schema_name, volume_name = parts
    return f"/Volumes/{catalog_name}/{schema_name}/{volume_name}"


class RotatingFileNamer:

    def __call__(self, default_name: str) -> str:
        name_parts = default_name.split(".")
        out_parts = name_parts[:-1]
        out_parts.append(datetime.now().strftime("%Y%m%d_%H%M%S_%f"))
        return ".".join(out_parts)


class FileRotator:

    def __call__(self, src: str, dst: str) -> None:
        if os.path.exists(src):
            os.rename(src, dst)


class VolumeRotator:

    def __init__(
        self,
        volume_path: str,
        databricks_host: Optional[str],
        databricks_token: Optional[str],
    ) -> None:

        self._volume_path = volume_path
        self._workspace_client: WorkspaceClient = WorkspaceClient(
            host=databricks_host, token=databricks_token
        )
        logger.info(f"Creating directory: {self._volume_path}")
        self._workspace_client.files.create_directory(self._volume_path)

    def __call__(self, src: str, dst: str) -> None:
        if os.path.exists(src):
            os.rename(src, dst)
            with open(dst, "r") as fin:
                content: str = fin.read()
                filename: str = Path(dst).name
                self._upload(
                    f"{self._volume_path}/{filename}.jsonl", content, overwrite=True
                )

    @retry(wait=wait_fixed(1), stop=stop_after_attempt(3))
    def _upload(
        self, file_path: str, contents: bytearray, overwrite: bool = True
    ) -> None:
        try:
            logger.info("Uploading log file", file_path=file_path)
            self._workspace_client.files.upload(file_path, contents, overwrite=True)
        except Exception as e:
            logger.error("Failed to upload file", file_path=file_path, error=str(e))


def _get_databricks_host_creds(
    databricks_host: Optional[str], databricks_token: Optional[str]
) -> Tuple[str, str]:
    if databricks_host is None or databricks_token is None:
        try:
            from mlflow.utils.databricks_utils import get_databricks_host_creds

            databricks_host = os.environ.get(
                "DATABRICKS_HOST", get_databricks_host_creds().host
            )
            databricks_token = os.environ.get(
                "DATABRICKS_TOKEN", get_databricks_host_creds().token
            )
        except Exception as e:
            databricks_host = os.environ.get("DATABRICKS_HOST", None)
            databricks_token = os.environ.get("DATABRICKS_TOKEN", None)
            logger.info("Failed to get databricks creds", error=str(e))

    return databricks_host, databricks_token


def create_rotator(
    volume_path: Optional[str],
    databricks_host: Optional[str],
    databricks_token: Optional[str],
) -> Callable[[str, str], None]:

    databricks_host, databricks_token = _get_databricks_host_creds(
        databricks_host, databricks_token
    )

    rotator: Callable[[str, str], None] = FileRotator()
    if all([v is not None for v in [databricks_host, databricks_token, volume_path]]):
        try:
            rotator = VolumeRotator(
                volume_path=volume_path,
                databricks_host=databricks_host,
                databricks_token=databricks_token,
            )
        except Exception as e:
            logger.error("Failed to create VolumeRotator", error=str(e))

    logger.info("Rotator created", rotator=type(rotator))
    return rotator


class SizeAndTimedRotatingVolumeHandler(TimedRotatingFileHandler):

    def __init__(
        self,
        filename: str,
        archive_path: Optional[str] = None,
        databricks_host: Optional[str] = None,
        databricks_token: Optional[str] = None,
        when: str = "h",
        interval: int = 1,
        max_bytes: int = 0,
        backup_count: int = 0,
        encoding: Optional[str] = None,
        delay: bool = False,
        utc: bool = False,
        at_time=None,
        errors=None,
    ):
        if max_bytes > 0:
            max_bytes = max(1024, max_bytes)
        self._max_bytes = max_bytes

        parent: Path = Path(filename).parent
        if not parent.exists():
            parent.mkdir(parents=True, exist_ok=True)

        TimedRotatingFileHandler.__init__(
            self,
            filename,
            when=when,
            interval=interval,
            backupCount=backup_count,
            encoding=encoding,
            delay=delay,
            utc=utc,
            atTime=at_time,
            errors=errors,
        )
        self.backup_count = backup_count

        self.namer: Callable[[str], str] = RotatingFileNamer()
        self.rotator: Callable[[str, str], None] = create_rotator(
            archive_path, databricks_host, databricks_token
        )

    def shouldRolloverOnSize(self) -> bool:
        if os.path.exists(self.baseFilename) and not os.path.isfile(self.baseFilename):
            return False

        if self.stream is None:
            return False

        if self._max_bytes > 0:
            self.stream.seek(0, 2)
            if self.stream.tell() >= self._max_bytes:
                return True
        return False

    def shouldRollover(self, record: bytearray) -> bool:
        return self.shouldRolloverOnSize() or super().shouldRollover(record)

    def getFilesToDelete(self) -> List[str]:
        dir_name, base_name = os.path.split(self.baseFilename)
        files_dict = {}
        for fileName in os.listdir(dir_name):
            _, ext = os.path.splitext(fileName)
            date_str = ext.replace(".", "")
            try:
                d = datetime.strptime(date_str, "%Y%m%d_%H%M%S_%f")
                files_dict[d] = fileName
            except:
                pass
        if len(files_dict) < self.backup_count:
            return []

        sorted_dict = dict(sorted(files_dict.items(), reverse=True))
        return [os.path.join(dir_name, v) for k, v in sorted_dict.items()][
            self.backup_count :
        ]


def rotating_volume_handler(
    filename: str,
    archive_path: Optional[str] = None,
    databricks_host: Optional[str] = None,
    databricks_token: Optional[str] = None,
    when: str = "h",
    interval: int = 1,
    max_bytes: int = 0,
    backup_count: int = 0,
    encoding: Optional[str] = None,
    delay: bool = False,
    utc: bool = False,
    at_time=None,
    errors=None,
) -> logging.Handler:
    return SizeAndTimedRotatingVolumeHandler(
        filename=filename,
        archive_path=archive_path,
        databricks_host=databricks_host,
        databricks_token=databricks_token,
        when=when,
        interval=interval,
        max_bytes=max_bytes,
        backup_count=backup_count,
        encoding=encoding,
        delay=delay,
        utc=utc,
        at_time=at_time,
        errors=errors,
    )
